#############
# $Id: Cluster_Density.pm 13782 2011-08-01 14:48:20Z ajb $
# Created By: gq1
# Last Maintained By: $Author: ajb $
# Created On: 2010-01-25
# Last Changed On: $Date: 2011-08-01 15:48:20 +0100 (Mon, 01 Aug 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-qc/tags/release-38.1/lib/npg_qc/api/loader/Cluster_Density.pm $

package npg_qc::api::loader::Cluster_Density;
use strict;
use warnings;
use Moose;
use Carp qw{carp cluck croak confess};
use English qw{-no_match_vars};
use Readonly;

extends qw{npg_qc::api::loader::base};

Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 13782 $ =~ /(\d+)/mxs; $r; };

## no critic (Documentation::RequirePodAtEnd)

=head1 NAME

npg_qc::api::loader::Cluster_Density

=head1 VERSION

$Revision: 13782 $

=head1 SYNOPSIS

  my $oClusterDensity = npg_qc::api::loader::Cluster_Density->new({
    id_run => $iIdRun,
    verbose => 1,
    run_folder => $sRunFolder,
  });

id_run or run_folder must be provided, or at least a path for npg_common::roles::run::path_info
which should expose the run_folder. verbose turns on the optional logs

=head1 DESCRIPTION

This loader object is responsible for parsing the Illumina cluster density by lane xml files and store into database

MooseX::Getopt is consumed by this object, so you can create the object with command line options

=head1 SUBROUTINES/METHODS

=head2 raw_xml_file

raw cluster density by lane xml file name
 
=cut
has 'raw_xml_file' => (isa           => q{Str},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{file name for raw cluster density xml},
                      );

sub _build_raw_xml_file {
  my $self = shift;
  return $self->reports_path().q{/NumClusters By Lane.xml};
}

=head2 pf_xml_file

pf cluster density by lane xml file name
 
=cut
has 'pf_xml_file' =>  (isa           => q{Str},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{file name for pf cluster density xml},
                      );

sub _build_pf_xml_file {
  my $self = shift;
  return $self->reports_path().q{/NumClusters By Lane PF.xml};
}

=head2 run

method to call, which is responsible for organising the loading of the cluster density data

  eval {
    $oIlluminaSummary->run();
  } or do {
    ... Your error handling here ...
  };

=cut

sub run {
  my ($self) = @_;

  $self->log(q{Loading Illumina cluster density data for Run } . $self->id_run() . q{ into QC database});

  $self->SUPER::run();

  my $dbh = $self->util->dbh();

  my $cluster_density_by_lane    = $self->parsing_xml($self->raw_xml_file());
  my $pf_cluster_density_by_lane = $self->parsing_xml($self->pf_xml_file());
  eval{
    $self->save_to_db_list($cluster_density_by_lane, 0, $dbh);
    $self->save_to_db_list($pf_cluster_density_by_lane, 1, $dbh);
    $dbh->commit;
    1;
  } or do {
    $dbh->rollback();
    croak q{cluster density data saving failed - } . $self->id_run().$EVAL_ERROR;
  };

  return 1;
}

=head2 parsing_xml

given one cluster dessity xml file, return a hashref

=cut

sub parsing_xml {
  my ($self, $xml) = @_;

  my $cluster_density_by_lane = {};

  my $lib_xml_parser = $self->util->parser();
  my $xml_dom = $lib_xml_parser->parse_file( $xml );
  my @lane_list = $xml_dom->getElementsByTagName('Lane');

  foreach my $lane (@lane_list){

    my $key = $lane->getAttribute('key');
    my $min = $lane->getAttribute('min');
    my $max = $lane->getAttribute('max');
    my $p50 = $lane->getAttribute('p50');
    $cluster_density_by_lane->{$key}->{min} = $min;
    $cluster_density_by_lane->{$key}->{max} = $max;
    $cluster_density_by_lane->{$key}->{p50} = $p50;
  }

  return $cluster_density_by_lane;
}

=head2 save_to_db_list

given a hash list and whether this list is for pf cluster, save them to database

=cut
sub save_to_db_list{
  my ($self, $cluster_density_by_lane, $is_pf, $dbh) = @_;

  foreach my $lane (keys %{$cluster_density_by_lane}){
    my $lane_values = $cluster_density_by_lane->{$lane};
    $self->save_to_db({lane => $lane,
                       is_pf=> $is_pf,
                       min  => $lane_values->{min},
                       max  => $lane_values->{max},
                       p50  => $lane_values->{p50},
                       dbh => $dbh,
                     });
  }

  return 1;
}

=head2 save_to_db

given one row of data, save it into database

=cut
sub save_to_db{
  my ($self, $args_ref) = @_;

  my $id_run = $self->id_run();
  my $lane   = $args_ref->{lane};
  my $is_pf  = $args_ref->{is_pf};
  my $min    = $args_ref->{min};
  my $max    = $args_ref->{max};
  my $p50    = $args_ref->{p50};

  my $dbh    = $args_ref->{dbh};
  if(!$dbh){
    $dbh = $self->util->dbh();
  }

  my $query = q{INSERT INTO cluster_density (id_run, position, is_pf, min, max, p50)
                VALUES (?, ?, ?, ?, ?, ?)
                ON DUPLICATE KEY UPDATE
                id_run = ?,
                position = ?,
                is_pf = ?,
                min = ?,
                max = ?,
                p50 = ?
                };

  eval{
    my $sth1 = $dbh->prepare_cached($query);
    $sth1->execute($id_run, $lane, $is_pf, $min, $max, $p50,
                   $id_run, $lane, $is_pf, $min, $max, $p50);
    #$dbh->commit;
    1;
  } or do {
    $dbh->rollback();
    croak "saving cluster density failed: $id_run - $lane: $EVAL_ERROR";
  };

  return 1;
}

no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=item MooseX::Getopt

=item npg_qc::api::loader::base

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2009 Guoying Qi (gq1@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
