#########
# Author:        gq1
# Maintainer:    $Author: ajb $
# Created:       2010-02-09
# Last Modified: $Date: 2011-08-01 15:48:20 +0100 (Mon, 01 Aug 2011) $
# Id:            $Id: Gerald_Summary.pm 13782 2011-08-01 14:48:20Z ajb $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-qc/tags/release-38.1/lib/npg_qc/api/loader/Gerald_Summary.pm $
#
package npg_qc::api::loader::Gerald_Summary;
use strict;
use warnings;
use Moose;
use Carp qw{carp cluck croak confess};
use English qw{-no_match_vars};
use Readonly;

extends qw{npg_qc::api::loader::base};

Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 13782 $ =~ /(\d+)/mxs; $r; };

Readonly::Scalar our $LANE_QC_FIELDS_TO_LOAD => {averageAlignScorePF      => 'av_alignment_score_pf',
                                                 errorPF                  => 'perc_error_rate_pf',
                                                 percentUniquelyAlignedPF => 'perc_align_pf',
                                                };

Readonly::Scalar our $LANE_SUMMARY_FIELDS_WITH_ERR_TO_LOAD
                                             => {averageAlignScorePF      => 'align_score_pf',
                                                 errorPF                  => 'error_rate_pf',
                                                 percentUniquelyAlignedPF => 'perc_align_pf',
                                                };
Readonly::Scalar our $LANE_EXPANDED_SUMMARY_FIELDS_NO_ERR_TO_LOAD
                                             => {errorRaw                 => 'perc_error_rate_raw',
                                                 infoContentRaw           => 'equiv_perfect_clusters_raw',
                                                 percentUniquelyAlignedPF => 'perc_align_pf2',
                                                 errorPF                  => 'perc_error_rate_pf',
                                                 infoContentPF            => 'equiv_perfect_clusters_pf'
                                                };
## no critic (Documentation::RequirePodAtEnd)

=head1 NAME

npg_qc::api::loader::Gerald_Summary

=head1 VERSION

$Revision: 13782 $

=head1 SYNOPSIS

  my $oGeraldSummary = npg_qc::api::loader::Gerald_Summary->new({
    id_run => $iIdRun,
    verbose => 1,
    run_folder => $sRunFolder,
  });

id_run or run_folder must be provided, or at least a path for npg_common::roles::run::path_info
which should expose the run_folder. verbose turns on the optional logs

=head1 DESCRIPTION

=head1 SUBROUTINES/METHODS

=head2 xml_file

bustard summary xml file name
 
=cut
has 'xml_file'     => (isa           => q{Str},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{file name for gerald summary xml},
                      );

sub _build_xml_file {
  my $self = shift;
  return $self->recalibrated_path().q{/Summary.xml};
}

=head2 run

method to call, which is responsible for organising the loading of the tile based error information into database

  eval {
    $oGeraldSummary->run();
  } or do {
    ... Your error handling here ...
  };

=cut

sub run {
  my ($self) = @_;

  $self->log(q{Loading Illumina Gerald Summary data for Run } . $self->id_run() . q{ into QC database});
  my $xml_file_name = $self->xml_file();
  if(! -e $xml_file_name){
     $self->log(qq{Gerald Summary XML not exist!!!\n$xml_file_name});
     return 1;
  }
  $self->_parsing_xml($xml_file_name);
  return 1;
}

=head2 parsing_xml

parsing xml to get the required data and store them into the database

=cut

sub _parsing_xml {
  my ($self, $xml) = @_;

  my $lib_xml_parser = $self->util->parser();
  my $xml_dom = $lib_xml_parser->parse_file( $xml );

  #processing lane summary and expanded summary 
  $self->log('Processing lane summary data for analysis and analysis_lane');
  my $lane_result_summary = $xml_dom->getElementsByTagName('LaneResultsSummary')->[0];
  my $expanded_lane_summary = $xml_dom->getElementsByTagName('ExpandedLaneSummary')->[0];
  my $lane_parameter_summary = $xml_dom->getElementsByTagName('LaneParameterSummary')->[0];
  $self->_process_lane_summary($lane_result_summary, $expanded_lane_summary, $lane_parameter_summary);

  #get errors by tile by lane and read
  $self->log('Processing error rate by tile data');
  my $tile_error_by_lane_element = $xml_dom->getElementsByTagName('TileErrorsByLane')->[0];

  my $lane_list = $tile_error_by_lane_element->getElementsByTagName('Lane');

  foreach my $lane (@{$lane_list}){
    $self->_process_error_by_lane($lane);
  }

  #commit all the data for this run to the database 
  $self->schema->txn_commit();

  return 1;
}

sub _process_lane_summary{
  my ($self, $lane_result_summary, $expanded_lane_summary, $lane_parameter_summary) = @_;

  my $lane_summary_db = {};
  my $tile_num_by_lane = {};

  $self->log('Processing lane results summary data');
  my $lane_result_summary_by_read = $lane_result_summary->getElementsByTagName('Read');
  foreach my $read (@{$lane_result_summary_by_read}){
     $self->_process_lane_result_summary_by_read($read , $lane_summary_db);
  }

  $self->log('Processing expanded lane summary data');
  my $expanded_lane_summary_by_read = $expanded_lane_summary->getElementsByTagName('Read');
  foreach my $read (@{$expanded_lane_summary_by_read}){
     $self->_process_expanded_lane_summary_by_read($read , $lane_summary_db);
  }

  $self->log('Processing lane parameter summary to get tile number per lane');
  my $lane_parameter_summary_one_lane = $lane_parameter_summary->getElementsByTagName('Lane');
  foreach my $lane (@{$lane_parameter_summary_one_lane}){
     my $lane_number = $lane->getElementsByTagName('laneNumber')->[0]->textContent();
     my $tile_number = $lane->getElementsByTagName('tileCountRaw')->[0]->textContent();
     $tile_num_by_lane->{$lane_number} = $tile_number;
  }

  $self->log('Populating lane summary data into database');
  foreach my $read (keys %{$lane_summary_db} ){

     my $id_analysis = $self->get_id_analysis($read);

     my $lane_summary_by_read = $lane_summary_db->{$read};

     foreach my $lane (keys %{$lane_summary_by_read}){

       my $lane_summary_by_read_lane = $lane_summary_by_read->{$lane};

       $lane_summary_by_read_lane->{id_run} = $self->id_run();
       $lane_summary_by_read_lane->{end}    = $read;
       $lane_summary_by_read_lane->{position} = $lane;
       $lane_summary_by_read_lane->{id_analysis} = $id_analysis;
       $lane_summary_by_read_lane->{tile_count} = $tile_num_by_lane->{$lane};

       $self->schema->resultset('AnalysisLane')->update_or_create(
                                         $lane_summary_by_read_lane
                                      );

     }
  }
  return 1;
}

sub _process_lane_result_summary_by_read{
  my ($self, $read, $lane_summary_db) = @_;

  my $read_number = $read->getElementsByTagName('readNumber')->[0]->textContent();

  $read_number = $self->transfer_read_number($read_number);

  $self->log("Processing read $read_number");

  my $lane_result_summary_by_read_lane = $read->getElementsByTagName('Lane');
  foreach my $lane (@{$lane_result_summary_by_read_lane}){
     $self->_process_lane_result_summary_by_read_lane($lane, $read_number, $lane_summary_db);
  }
  return 1;
}

sub _process_lane_result_summary_by_read_lane{
  my ( $self, $lane, $read_number, $lane_summary_db ) = @_;

  my $lane_number = $lane->getElementsByTagName('laneNumber')->[0]->textContent();
  my $lane_yield = $lane->getElementsByTagName('laneYield');

  $self->log("Processing read $read_number lane $lane_number");

  foreach my $field ( keys %{ $LANE_SUMMARY_FIELDS_WITH_ERR_TO_LOAD } ){

    my $db_field = $LANE_SUMMARY_FIELDS_WITH_ERR_TO_LOAD->{$field};
    my $element = $lane->getElementsByTagName($field)->[0];
    if($element){
      my $mean  = $element->getElementsByTagName('mean')->[0]->textContent();
      my $stdev = $element->getElementsByTagName('stdev')->[0]->textContent();
      $lane_summary_db->{$read_number}->{$lane_number}->{$db_field} = $mean;
      $lane_summary_db->{$read_number}->{$lane_number}->{$db_field.q{_err}} = $stdev;
    }
  }
  return 1;
}

sub _process_expanded_lane_summary_by_read{
  my ($self, $read, $lane_summary_db) = @_;

  my $read_number = $read->getElementsByTagName('readNumber')->[0]->textContent();

  $read_number = $self->transfer_read_number($read_number);

  $self->log("Processing read $read_number");

  my $expanded_lane_summary_by_read_lane = $read->getElementsByTagName('Lane');
  foreach my $lane (@{$expanded_lane_summary_by_read_lane}){
     $self->_process_expanded_lane_summary_by_read_lane($lane, $read_number, $lane_summary_db);
  }
  return 1;
}


sub _process_expanded_lane_summary_by_read_lane{
  my ( $self, $lane, $read_number, $lane_summary_db ) = @_;

  my $lane_number = $lane->getElementsByTagName('laneNumber')->[0]->textContent();

  $self->log("Processing read $read_number lane $lane_number");

  my $expanded_lane_summary_by_read_lane = {};

  foreach my $field ( keys %{ $LANE_EXPANDED_SUMMARY_FIELDS_NO_ERR_TO_LOAD } ){

    my $db_field = $LANE_EXPANDED_SUMMARY_FIELDS_NO_ERR_TO_LOAD ->{$field};
    my $element = $lane->getElementsByTagName($field)->[0];
    if($element){
      my $mean  = $element->getElementsByTagName('mean')->[0]->textContent();
      $lane_summary_db->{$read_number}->{$lane_number}->{$db_field} = $mean;
    }
  }
  return 1;
}

sub _process_error_by_lane {
  my ($self, $lane) = @_;

  my $lane_number = $lane->getElementsByTagName('laneNumber')->[0]->textContent();

  $self->log(qq{Processing lane $lane_number});

  my $read_list = $lane->getElementsByTagName('Read');

  foreach my $read (@{$read_list}){
    $self->_process_error_by_read_lane($lane_number, $read);
  }

  return 1;
}

sub _process_error_by_read_lane{
  my ($self, $lane_number, $read) = @_;

  my $read_number = $read->getElementsByTagName('readNumber')->[0]->textContent();

  $read_number = $self->transfer_read_number($read_number);

  $self->log(qq{Processing Lane $lane_number Read $read_number});

  my $tile_list = $read->getElementsByTagName('Tile');

  foreach my $tile (@{$tile_list}){
    $self->_process_error_by_tile_read_lane($lane_number, $read_number, $tile);
  }
  return 1;
}

sub _process_error_by_tile_read_lane{
  my ($self, $lane_number, $read_number, $tile) = @_;

  #get tile number  
  my $tile_number = $tile->getElementsByTagName('tileNumber')->[0]->textContent();
  $tile_number += 0;

  #get id_run_tile number, if not available insert one
  my $id_run_tile = $self->get_id_run_tile($lane_number, $read_number, $tile_number);

  #process one tile data into lane_qc table
  my $lane_qc_db ={end         => $read_number,
                   id_run_tile => $id_run_tile,
                  };

  foreach my $field (keys %{$LANE_QC_FIELDS_TO_LOAD}){

    my $element_list = $tile->getElementsByTagName($field);
    if($element_list){

      my $value = $element_list->[0]->textContent();
      $lane_qc_db->{$LANE_QC_FIELDS_TO_LOAD->{$field}} = $value;
    }
  }

  if(! defined $lane_qc_db->{av_alignment_score_pf} && ! defined $lane_qc_db->{perc_align_pf} ){
    $lane_qc_db->{perc_error_rate_pf} = undef;
  }

  eval{
    my $schema = $self->schema();
    my $lane_qc_obj = $schema->resultset('LaneQc')->update_or_create(
                                         $lane_qc_db
                                      );
    #$schema->txn_commit();
    1;
  } or do{
    croak $EVAL_ERROR;
  };
  return 1;
}

no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=item MooseX::Getopt

=item npg_qc::api::loader::base

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2009 Guoying Qi (gq1@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
