#############
# $Id: Illumina_Analysis.pm 13782 2011-08-01 14:48:20Z ajb $
# Created By: ajb
# Mast Maintained By: $Author: ajb $
# Created On: 2009-11-24
# Last Changed On: $Date: 2011-08-01 15:48:20 +0100 (Mon, 01 Aug 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-qc/tags/release-38.1/lib/npg_qc/api/loader/Illumina_Analysis.pm $

package npg_qc::api::loader::Illumina_Analysis;
use strict;
use warnings;
use Moose;
use Carp qw{carp cluck croak confess};
use English qw{-no_match_vars};
use Readonly;

# these are loader modules which know the loading commands to dump data into the database
use npg_qc::api::loader::Fastqcheck;
use npg_qc::api::loader::Matrix;
use npg_qc::api::loader::Offset;
use npg_qc::api::loader::Bustard_Summary;
use npg_qc::api::loader::Gerald_Summary;
use npg_qc::api::loader::Qcal_Report;
use npg_qc::api::loader::Tile_Score;
use npg_qc::api::loader::Run_Config;
use npg_qc::api::loader::Signal_Mean;

# this is responsible for submitting a request to cache some data up for speed on the website
use npg_qc::api::run_caching;

extends qw{npg_qc::api::loader::base};

Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 13782 $ =~ /(\d+)/mxs; $r; };

## no critic (Documentation::RequirePodAtEnd)

=head1 NAME

npg_qc::api::loader::Illumina_Analysis

=head1 VERSION

$Revision: 13782 $

=head1 SYNOPSIS

  my $oIlluminaAnalysisLoader = npg_qc::api::loader::Illumina_Analysis->new({
    id_run => $iIdRun,
  });

  my $oIlluminaAnalysisLoader = npg_qc::api::loader::Illumina_Analysis->new({
    run_folder => $sRunFolder,
  });

=head1 DESCRIPTION

This loader object is responsible for loading into the npg-qc database the illumina produced statistics, plus some other stats
which have been generated during the post_qseq part of the npg-pipeline, although not the in-house qc statistics

MooseX::Getopt is consumed by this object, so you can create the object with command line options

=head1 SUBROUTINES/METHODS

=head2 run

The main method to be called which knows what to do, and will process through

  eval {
    $oIlluminaAnalysisLoader->run();
  } or do {
    ... Your error handling here ...
  };

=cut

sub run {
  my ($self) = @_;

  $self->log(q{Loading Illimina Analysis Data for Run } . $self->id_run() . q{ into QC database});

  $self->SUPER::run();

  my $schema = $self->util->schema();

  #loading chip summary, tile base result and lane summary from bustard summary xml file
  my $bustard_summary_loader = npg_qc::api::loader::Bustard_Summary->new({
                            runfolder_path => $self->runfolder_path(),
                            schema         => $schema,
  });
  $bustard_summary_loader->run();

  #loading extra data for tile based and lane based data from gerald summary xml file
  my $gerald_summary_loader = npg_qc::api::loader::Gerald_Summary->new({
                            runfolder_path => $self->runfolder_path(),
                            schema         => $schema,
  });
  $gerald_summary_loader->run();

  #loading qcalreport files into analysis_lane_qcal table
  my $qcalreport_loader = npg_qc::api::loader::Qcal_Report->new({
                            runfolder_path => $self->runfolder_path(),
                            schema         => $schema,
  });
  $qcalreport_loader->run();

  #loading q values into qcal table
  my $fastqcheck_loader = npg_qc::api::loader::Fastqcheck->new({
                            runfolder_path => $self->runfolder_path(),
                            schema         => $schema,
  });
  $fastqcheck_loader->run();

  #loading signal mean into signal mean table
  my $signal_mean_loader = npg_qc::api::loader::Signal_Mean->new({
                            runfolder_path => $self->runfolder_path(),
                            schema         => $schema,
  });
  $signal_mean_loader->run();

  #loading run cofig into run_config table
  my $run_config_loader = npg_qc::api::loader::Run_Config->new({
                            runfolder_path => $self->runfolder_path(),
                            schema         => $schema,
  });
  $run_config_loader->run();

  #loading tile score into tile_score table
  my $tile_score_loader = npg_qc::api::loader::Tile_Score->new({
                            runfolder_path => $self->runfolder_path(),
                            schema         => $schema,
  });
  $tile_score_loader->run();

  #loading matrix into frequency_response_matrix  table
  my $matrix_loader = npg_qc::api::loader::Matrix->new({
                            runfolder_path => $self->runfolder_path(),
                            schema         => $schema,
  });
  $matrix_loader->run();

  #loading offset files into offset table
  #my $offset_loader = npg_qc::api::loader::Offset->new({
  #                          runfolder_path => $self->runfolder_path(),
  #                          schema         => $schema,
  #});
  #$offset_loader->run();

  $self->log(q{All QC data loading finished for run }.$self->id_run());

  $self->log(q{Caching data});
  $self->_run_caching();
  $self->log(q{Sending caching finished});

  return 1;
}

has q{run_cache_object} => (isa => q{npg_qc::api::run_caching}, is => q{ro}, lazy_build => 1, init_arg => undef);
sub _build_run_cache_object {
  my ($self) = @_;
  return npg_qc::api::run_caching->new({util => $self->util()});
}

sub _run_caching{
  my ($self) = @_;

  my $id_run = $self->id_run();

  my $api_run_caching = $self->run_cache_object();

  $self->log(q{caching lane summary data and run graph});
  eval {
    $api_run_caching->cache_query_lane_summary($id_run, 1, 1);
    $api_run_caching->run_graph($id_run, 1, 1);
    $api_run_caching->instrument_statistics($id_run, 1);
    if ($self->is_paired_read()) {
      $api_run_caching->cache_query_lane_summary($id_run, 2, 1);
      $api_run_caching->run_graph($id_run, 2, 1);
      $api_run_caching->instrument_statistics($id_run, 2);
    }
    1;
  } or do {
    croak $EVAL_ERROR;
  };

  return;
}

no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=item MooseX::Getopt

=item npg_common::roles::run::short_info

=item npg_common::roles::run::path_info

=item npg_common::roles::log

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2009 Andy Brown (setitesuk@gmail.com)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
