#########
# Author:        gq1
# Maintainer:    $Author: ajb $
# Created:       2010-02-16
# Last Modified: $Date: 2011-08-01 15:48:20 +0100 (Mon, 01 Aug 2011) $
# Id:            $Id: Qcal_Report.pm 13782 2011-08-01 14:48:20Z ajb $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-qc/tags/release-38.1/lib/npg_qc/api/loader/Qcal_Report.pm $
#
package npg_qc::api::loader::Qcal_Report;
use strict;
use warnings;
use Moose;
use Carp qw{carp cluck croak confess};
use English qw{-no_match_vars};
use Readonly;

extends qw{npg_qc::api::loader::base};

Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 13782 $ =~ /(\d+)/mxs; $r; };

Readonly::Scalar our $VALUE_FOR_DATA_MINING_EFFICIENCY => -1;

## no critic (Documentation::RequirePodAtEnd)

=head1 NAME

npg_qc::api::loader::Qcal_Report

=head1 VERSION

$Revision: 13782 $

=head1 SYNOPSIS

  my $oQcalReport = npg_qc::api::loader::Qcal_Report->new({
    id_run => $iIdRun,
    verbose => 1,
    run_folder => $sRunFolder,
  });

id_run or run_folder must be provided, or at least a path for npg_common::roles::run::path_info
which should expose the run_folder. verbose turns on the optional logs

=head1 DESCRIPTION

=head1 SUBROUTINES/METHODS

=head2 run

method to call, which is responsible for organising the loading of the tile based cluster data into database

  eval {
    $oQcalReport->run();
  } or do {
    ... Your error handling here ...
  };

=cut

sub run {
  my ($self) = @_;

  $self->log(q{Loading Illumina qcalreport data for Run } . $self->id_run() . q{ into QC database});
  $self->_process_run_qcalreports($self->recalibrated_path());
  return 1;
}

sub _process_run_qcalreports {
  my ($self, $dtfolder) = @_;

  my $dbh = $self->util->dbh();

  eval {
    opendir my $dh, $dtfolder or ( carp "Cannot open $dtfolder" && return);
    my @qcalresults = grep { /_\d+_qcalreport\.txt/mxs } sort readdir $dh;
    closedir $dh or croak qq(Cannot close '$dtfolder');

    $self->log(q(There are ).@qcalresults.q( qcalreport.txt files to process!));
  #########
  # Now process per-lane s_X_qcalreport.txt files
  #
    for my $qcal (@qcalresults) {

      my ($raw, $filtered) = $self->_process_qcal("$dtfolder/$qcal");

      my ($lane_position, $end);
      if ($qcal =~ /s_\d+_qcalreport/mxs) {
        ($lane_position)  = $qcal =~ /s_(\d+)_qcalreport/mxs;
	$end = 1;
      } else {
        ($lane_position, $end)  = $qcal =~ /s_(\d+)_(\d+)_qcalreport/mxs;
      }

      if(!$end){
         $end = 1;
      }

      my $id_analysis_lane = $self->_get_id_analysis_lane($end, $lane_position);

      if(!$id_analysis_lane) {
        $self->log("$lane_position for '$dtfolder/$qcal' has no id_analysis_lane! Skipping!");
        next;
      }

      for my $row (@{$raw}) {
        $self->_update_one_row($row, $id_analysis_lane  );
      }

      for my $row (@{$filtered}) {
        $self->_update_one_row($row, $id_analysis_lane );
      }
      $self->schema->txn_commit();
    }
    1;
  } or do {
   croak $EVAL_ERROR;
  };

  return 1;
}

sub _process_qcal {
  my ($self, $fn) = @_;

  $self->log("Processing $fn");
  my ($dtfn) = $fn =~ m{([[:lower:]\d_\-\./]+)}mixs;

  local $RS = "\n\n";
  open my $fh, q(<), $dtfn or croak qq(Cannot open '$dtfn') ;
  my $rev      = <$fh>;
  my $tilelist = <$fh>;
  #my $tiles    = <$fh>;
  my $raw      = $self->_read_qcal_table($fh);
  my $filtered = $self->_read_qcal_table($fh);
  close $fh or croak "Closing $dtfn: $ERRNO";

#  return ($tile_count, $raw, $filtered);
  return ($raw, $filtered);
}

sub _read_qcal_table {
  my ($self, $fh) = @_;

  local $RS = "\n";
  my ($tmp);
  my $chastity = 0;

  while(my $tmp = <$fh>) { ## no critic (Variables::ProhibitReusedNames)
    #########
    # read up to table header
    #
    if($tmp =~ /Results[ ]for[ ](raw|quality[ ]filtered)?[ ]?data/mxs) {
      if($tmp =~ /raw/mxs) {
	      $chastity = $VALUE_FOR_DATA_MINING_EFFICIENCY; # could be undef here but data-mining selects on NULL are inefficient later on
      }

      if($tmp =~ /CHASTITY/msx) {
        ($chastity) = $tmp =~ /CHASTITY>=([\d\.]+)/mxs;
      }
      last;
    }
  }

  $tmp = <$fh>; # - - - - - -
  $tmp = <$fh>; # QV      this QV or higher               |       this QV only                    |       expected
  $tmp = <$fh>; #         # error # bases % error % total |       # error # bases % error % total |       % error
  $tmp = <$fh>; # -       -       -       -       -       -       -       -       -       -       -       -

  my $table = [];
  while(my $line = <$fh>) {
    if($line eq "\n") {
      last;
    }

    chomp $line;
    my ($qval,
	$cum_error, $cum_bases, $cum_perc_error, $cum_perc_total,
	$dummy_1,
	$error_rate, $base_count, $perc_error, $perc_total,
	$dummy_2,
	$exp_error) = map { (/\d+/mxs)?$_+0:$_; } split /\t/mxs, $line;

    push @{$table}, [$chastity, $qval,
		     $cum_error, $cum_bases, $cum_perc_error, $cum_perc_total,
		     $error_rate, $base_count, $perc_error, $perc_total,
		     $exp_error];
  }

  return $table;
}

sub _get_id_analysis_lane {
  my ($self, $end, $lane) = @_;

  my $id_analysis = $self->get_id_analysis($end);

  my $analysis_lane = $self->schema->resultset('AnalysisLane')
                        ->find_or_create(
                            {position  => $lane,
                             id_analysis => $id_analysis,
                            });
  return $analysis_lane->id_analysis_lane();
}

sub _update_one_row{
  my ($self, $row, $id_analysis_lane) = @_;

  my @fields
  = qw(chastity qv cum_error cum_bases cum_perc_error cum_perc_total error bases perc_error perc_total exp_perc_error);

  my %fields_db = ();

  @fields_db{@fields} = @{$row};

  $fields_db{id_analysis_lane} = $id_analysis_lane;

  $self->schema->resultset('AnalysisLaneQcal')->update_or_create(\%fields_db);

  return 1;
}

no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=item MooseX::Getopt

=item npg_qc::api::loader::base

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2009 Guoying Qi (gq1@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
