#########
# Author:        gq1
# Maintainer:    $Author: ajb $
# Created:       2010-02-19
# Last Modified: $Date: 2011-08-01 15:48:20 +0100 (Mon, 01 Aug 2011) $
# Id:            $Id: Tile_Score.pm 13782 2011-08-01 14:48:20Z ajb $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-qc/tags/release-38.1/lib/npg_qc/api/loader/Tile_Score.pm $
#
package npg_qc::api::loader::Tile_Score;
use strict;
use warnings;
use Moose;
use Carp qw{carp cluck croak confess};
use English qw{-no_match_vars};
use Readonly;
use Perl6::Slurp;

extends qw{npg_qc::api::loader::base};

Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 13782 $ =~ /(\d+)/mxs; $r; };
Readonly our $THREE                       => 3;
Readonly our $FOUR                        => 4;
Readonly our $FIVE                        => 5;

## no critic (Documentation::RequirePodAtEnd ProhibitComplexRegexes)

=head1 NAME

npg_qc::api::loader::Tile_Score

=head1 VERSION

$Revision: 13782 $

=head1 SYNOPSIS

=head1 DESCRIPTION

=head1 SUBROUTINES/METHODS

=head2 file_list

tile score and rescore file list
 
=cut
has 'file_list'    => (isa           => q{ArrayRef},
                       is            => q{rw},
                       lazy_build    => 1,
                       documentation => q{score and rescore file name list},
                      );

sub _build_file_list {
  my $self = shift;

  my @file_list = ();

  my $folder_path = $self->score_path();
  opendir my $dh, $folder_path or ( carp "Cannot open $folder_path" and return \@file_list);
  @file_list = map {$folder_path.qq{$_}} grep { /score[.]txt/mxs } sort readdir $dh;
  closedir $dh or croak qq(Cannot close $folder_path);

  return \@file_list;
}

=head2 run

=cut

sub run {
  my ($self) = @_;

  $self->log(q{Loading Illumina tile score files for Run } . $self->id_run() . q{ into QC database});

  my $file_list = $self->file_list();
  $self->log('There are '.scalar @{$file_list}. ' files to process');

  foreach my $tile_score_file (@{$file_list}){

    $self->_process_tile_score($tile_score_file);
  }

  $self->schema->txn_commit();

  return 1;
}

sub _process_tile_score {
    my ($self, $file_url) = @_;

    my $id_run = $self->id_run();

    my ($position, $tile) = $file_url =~ /s_(\d{1})_0*(\d+)_.*?[.]txt/xms;

    my $rescore = $file_url =~ /s_\d{1}_\d+_rescore.txt/xms ? 1
                :                                             0
                ;

    my $args_ref = { id_run   => $id_run,
                     position => $position,
                     tile     => $tile,
                     rescore  => $rescore,
                   };

    my $file = slurp $file_url;

    $self->log("position: $position, tile: $tile, rescore: $rescore");

    my $end      = 0;
    my @ends;
    if ($file !~ /bases[ ]of[ ]sequence[ ]found.*?bases[ ]of[ ]sequence[ ]found.*?bases[ ]of[ ]sequence[ ]found/xms) {

      push @ends, $file;
    } else {

      # for version 1.0
      if ($file =~ /RUN_TIME/xms) {
        @ends = split /RUN_TIME/xms, $file;
        shift @ends;

        foreach my $text (@ends) {
          $text = '#RUN_TIME'.$text;
        }
      } else { # for version rc1.3
        @ends = split /Lane/xms, $file;
        shift @ends;

        foreach my $text (@ends) {
          $text = '#Lane'.$text;
        }
      }

    }

    foreach my $text (@ends) {

      last if ($end == 2);
      $end++;

      $args_ref->{end} = $end;

      my $id_run_tile = $self->get_id_run_tile($position, $end, $tile);
      $args_ref->{id_run_tile} = $id_run_tile;

      my @lines = split /\n/xms, $text;
      my ($cumulative_errors_by_cycle, $log_likelihood, $information_content_by_cycle, $error_rate_relative_reference_cycle_nucleotide);

      foreach my $line (@lines) {
        my $processed = $line =~ /\A\!\d+/xms ? push @{$cumulative_errors_by_cycle}, $line
                      : $line =~ /\A\>\d+/xms ? push @{$log_likelihood}, $line
                      : $line =~ /\A\~\d+/xms ? push @{$information_content_by_cycle}, $line
                      : $line =~ /\A\@\d+/xms ? push @{$error_rate_relative_reference_cycle_nucleotide}, $line
                      :                         q{}
                      ;
      }

      $self->_tile_score($text, $args_ref);
      $self->_cumulative_errors_by_cycle($cumulative_errors_by_cycle, $args_ref);
      $self->_log_likelihood($log_likelihood, $args_ref);
      $self->_information_content_by_cycle($information_content_by_cycle, $args_ref);
      $self->_error_rate_relative_reference_cycle_nucleotide($error_rate_relative_reference_cycle_nucleotide, $args_ref);
      $self->_errors_by_cycle($text, $args_ref);
      $self->_error_rate_relative_to_reference_base_inc_blanks($text, $args_ref);
      $self->_error_rate_relative_to_reference_base_no_blanks($text, $args_ref);
      $self->_error_rate_relative_sequence_base($text, $args_ref);
      $self->_errors_by_nucleotide($text, $args_ref);
      $self->_errors_by_cycle_and_nucleotide($text, $args_ref);
      $self->_most_common_word($text, $args_ref);
      $self->_most_common_blank_pattern($text, $args_ref);
    }

    return 1;
  }

  sub _tile_score {
    my ($self, $file, $args_ref) = @_;

    $file =~ s/Breakdown[ ]of[ ]errors[ ]by[ ]cycle.*//gxms;

    my ($base_count)              = $file =~ /(\d+)\ bases\ of\ sequence\ found/xms;
    my ($error_count)             = $file =~ /(\d+)\ were\ errors/xms;
    my ($blank_count)             = $file =~ /(\d+)\ were\ blanks/xms;
    my ($unique_alignments)       = $file =~ /unique\ alignments\ :\ (\d+)/xms;
    my ($ua_total_score)          = $file =~ /total\ score\ (\d+)/xms;
    my ($cycles)                  = $file =~ /cycles\ :\ (\d+)/xms;
    my ($score_version)           = $file =~ /score\.pl\,v\ (\d+\.\d+)/xms;
    my ($phagealign_version)      = $file =~ /PhageAlign\.cpp\,v\ (\d+\.\d+)/xms;
    my ($max_blanks)              = $file =~ /\#MAX_BLANKS\ (\d+)/xms;
    my ($seq_length)              = $file =~ /\#SEQ_LENGTH\ (\d+)/xms;
    my ($genome_file)             = $file =~ /\#GENOME_FILE\ (.*?)\n/xms;
    my ($bases_used)              = $file =~ /\#BASES_USED\ (.*?)\n/xms;
    my ($qualityfilter_version)   = $file =~ /qualityFilter\.pl\,v\ (\d+\.\d+)/xms;
    my ($filter_criterion)        = $file =~ /\#FILTER_CRITERION\ \(\((.*?)\)\)/xms;

    my $date_run = {
      score         => q{},
      phagealign    => q{},
      qualityfilter => q{},
    };

    $file =~ s/\#MAX_BLANKS.*//gxms;
    my @lines = split /\n/xms, $file;

    for my $i (0..scalar@lines) {

      next if (!$lines[$i]);

      my $program = $lines[$i] =~ /PhageAlign/xms    ? q{phagealign}
                  : $lines[$i] =~ /score[.]pl/xms     ? q{score}
                  : $lines[$i] =~ /qualityFilter/xms ? q{qualityfilter}
                  :                                    next
                  ;

      my $date_row = $i - 1;
      my @temp = split q{ }, $lines[$date_row];
      my $months = { Jan => '01', Feb => '02', Mar => '03', Apr => '04', May => '05', Jun => '06', Jul => '07', Aug => '08', Sep => '09', Oct => '10', Nov => '11', Dec => '12' };
      $temp[3] = sprintf '%02d', $temp[3]; ## no critic (ProhibitMagicNumbers);
      $date_run->{$program} = $temp[-1] . q{-} . $months->{$temp[2]} . q{-} . $temp[3] . q{ } . $temp[4]; ## no critic (ProhibitMagicNumbers);

    }

    $genome_file            ||= q{};
    $seq_length             ||= q{};
    $max_blanks             ||= q{};
    $phagealign_version     ||= q{};
    $score_version          ||= q{};
    $bases_used             ||= q{};
    $qualityfilter_version  ||= q{};
    $filter_criterion       ||= q{};


    my $fields_db = {base_count             => $base_count,
                     error_count            => $error_count,
                     blank_count            => $blank_count,
                     unique_alignments      => $unique_alignments,
                     ua_total_score         => $ua_total_score,
                     cycles                 => $cycles,
                     score_version          => $score_version,
                     score_date_run         => $date_run->{score},
                     phagealign_version     => $phagealign_version,
                     phagealign_date_run    => $date_run->{phagealign},
                     max_blanks             => $max_blanks,
                     seq_length             => $seq_length,
                     genome_file            => $genome_file,
                     bases_used             => $bases_used,
                     qualityfilter_version  => $qualityfilter_version,
                     qualityfilter_date_run => $date_run->{qualityfilter},
                     filter_criterion       => $filter_criterion,
                     id_run_tile            => $args_ref->{id_run_tile},
                     rescore                => $args_ref->{rescore},
                    };
    $self->schema->resultset('TileScore')->update_or_create($fields_db);
    return 1;
  }

  sub _cumulative_errors_by_cycle {
    my ($self, $table_rows, $args_ref) = @_;

    foreach my $row (@{$table_rows}) {

      my @temp = split q{ }, $row;
      ($temp[0]) = $temp[0] =~ /(\d+)/xms;

      my $fields_db = { cycle       => $temp[0],
                        one         => $temp[1],
                        two         => $temp[2],
                        three       => $temp[3],## no critic (ProhibitMagicNumbers);
                        four        => $temp[4],## no critic (ProhibitMagicNumbers);
                        five        => $temp[5],## no critic (ProhibitMagicNumbers);
                        rescore     => $args_ref->{rescore},
                        id_run_tile =>$args_ref->{id_run_tile},
                       };
      $self->schema->resultset('CumulativeErrorsByCycle')->update_or_create($fields_db);
    }
    return 1;
  }

  sub _log_likelihood {
    my ($self, $table_rows, $args_ref) = @_;

    foreach my $row (@{$table_rows}) {

      my @temp = split q{ }, $row;
      ($temp[0]) = $temp[0] =~ /(\d+)/xms;

      my $fields_db = { cycle          => $temp[0],
                        read_as        => $temp[1],
                        really_a       => $temp[2],
                        really_c       => $temp[3],## no critic (ProhibitMagicNumbers);
                        really_g       => $temp[4],## no critic (ProhibitMagicNumbers);
                        really_t       => $temp[5],## no critic (ProhibitMagicNumbers);
                        rescore        => $args_ref->{rescore},
                        id_run_tile    =>$args_ref->{id_run_tile},
                       };
      $self->schema->resultset('LogLikelihood')->update_or_create($fields_db);
    }

    return 1;
  }

  sub _information_content_by_cycle {
    my ($self, $table_rows, $args_ref) = @_;

    foreach my $row (@{$table_rows}) {

      my @temp = split q{ }, $row;
      ($temp[0]) = $temp[0] =~ /(\d+)/xms;

      my $fields_db = {cycle        => $temp[0],
                       equiv_info   => $temp[1],
                       align        => $temp[2],
                       total        => $temp[3],## no critic (ProhibitMagicNumbers);
                       rescore      => $args_ref->{rescore},
                       id_run_tile  =>$args_ref->{id_run_tile},
                      };
     $self->schema->resultset('InformationContentByCycle')->update_or_create($fields_db);

    }

    return 1;
  }

  sub _error_rate_relative_reference_cycle_nucleotide {
    my ($self, $table_rows, $args_ref) = @_;

    foreach my $row (@{$table_rows}) {

      my @temp = split q{ }, $row;
      ($temp[0]) = $temp[0] =~ /(\d+)/xms;

      my $fields_db = {cycle      => $temp[0],
                       read_as    => $temp[1],
                       really_a   => $temp[2],
                       really_c   => $temp[3],## no critic (ProhibitMagicNumbers);
                       really_g   => $temp[4],## no critic (ProhibitMagicNumbers);
                       really_t   => $temp[5],## no critic (ProhibitMagicNumbers);
                       rescore    => $args_ref->{rescore},
                       id_run_tile=> $args_ref->{id_run_tile},
                      };
      $self->schema->resultset('ErrorRateRelativeReferenceCycleNucleotide')->update_or_create($fields_db);
    }

    return 1;
  }

  sub _errors_by_cycle {
    my ($self, $file, $args_ref) = @_;

    my ($wanted) = $file =~ /Breakdown[ ]of[ ]errors[ ]by[ ]cycle\n(.*?)\n\nError[ ]rate[ ]relative[ ]to[ ]reference[ ]base/xms;
    my @lines = split /\n/xms, $wanted;
    shift @lines;

    foreach my $line (@lines) {

      my @temp =  split /\s+/xms, $line;
      my $fields_db = {cycle            => $temp[0],
                       error_percentage => $temp[1],
                       blank_percentage => $temp[2],
                       rescore          => $args_ref->{rescore},
                       id_run_tile      => $args_ref->{id_run_tile},
                      };
      $self->schema->resultset('ErrorsByCycle')->update_or_create($fields_db);
    }

    return 1;
  }

  sub _error_rate_relative_to_reference_base_inc_blanks {
    my ($self, $file, $args_ref) = @_;

    my $xml = qq{<error_rate_reference_including_blanks>\n};

    my ($wanted) = $file =~ /Error[ ]rate[ ]relative[ ]to[ ]reference[ ]base[ ][(]including[ ]blanks[)]\n(.*?)\nError[ ]rate[ ]relative[ ]to[ ]reference[ ]base[ ][(]excluding[ ]blanks[)]/xms;
    my @lines = split /\n/xms, $wanted;
    splice @lines, 0, $FOUR;

    foreach my $line (@lines) {

      my @temp = split /\s+/xms, $line;

      my $fields_db = {really             => $temp[0],
                       read_as_blank      => $temp[1],
                       read_as_a          => $temp[2],
                       read_as_c          => $temp[3],## no critic (ProhibitMagicNumbers);
                       read_as_g          => $temp[4],## no critic (ProhibitMagicNumbers);
                       read_as_t          => $temp[5],## no critic (ProhibitMagicNumbers);
                       rescore            => $args_ref->{rescore},
                       id_run_tile        => $args_ref->{id_run_tile},
                      };
      $self->schema->resultset('ErrorRateReferenceIncludingBlank')->update_or_create($fields_db);
    }

    return 1;
  }

  sub _error_rate_relative_to_reference_base_no_blanks {
    my ($self, $file, $args_ref) = @_;

    my ($wanted) = $file =~ /Error[ ]rate[ ]relative[ ]to[ ]reference[ ]base[ ][(]excluding[ ]blanks[)]\n(.*?)\nError[ ]rate[ ]relative[ ]to[ ]sequenced[ ]base/xms;
    my @lines = split /\n/xms, $wanted;
    splice @lines, 0, $FOUR;

    foreach my $line (@lines) {

      my @temp = split /\s+/xms, $line;

      my $fields_db = {really             => $temp[0],
                       read_as_a          => $temp[1],
                       read_as_c          => $temp[2],
                       read_as_g          => $temp[3],## no critic (ProhibitMagicNumbers);
                       read_as_t          => $temp[4],## no critic (ProhibitMagicNumbers);
                       rescore            => $args_ref->{rescore},
                       id_run_tile        => $args_ref->{id_run_tile},
                      };
      $self->schema->resultset('ErrorRateReferenceNoBlank')->update_or_create($fields_db);

    }
    return 1;
  }

  sub _error_rate_relative_sequence_base {
    my ($self, $file, $args_ref) = @_;

    my ($wanted) = $file =~ /Error[ ]rate[ ]relative[ ]to[ ]sequenced[ ]base(.*?)\nBreakdown[ ]of[ ]errors[ ]by[ ]nucleotide/xms;
    my @lines = split /\n/xms, $wanted;
    splice @lines, 0, $FIVE;

    foreach my $line (@lines) {

      my @temp = split /\s+/xms, $line;

      my $fields_db = {read_as           => $temp[0],
                       really_a          => $temp[1],
                       really_c          => $temp[2],
                       really_g          => $temp[3],## no critic (ProhibitMagicNumbers);
                       really_t          => $temp[4],## no critic (ProhibitMagicNumbers);
                       rescore           => $args_ref->{rescore},
                       id_run_tile       => $args_ref->{id_run_tile},
                      };
      $self->schema->resultset('ErrorRateRelativeSequenceBase')->update_or_create($fields_db); }

    return 1;
  }

  sub _errors_by_nucleotide {
    my ($self, $file, $args_ref) = @_;

    my ($wanted) = $file =~ /Breakdown[ ]of[ ]errors[ ]by[ ]nucleotide\n(.*?)\nFull[ ]breakdown[ ]of[ ]errors[ ]by[ ]cycle[ ]and[ ]nucleotide/xms;
    my @lines = split /\n/xms, $wanted;
    splice @lines, 0, $THREE;

    foreach my $line (@lines) {

      my @temp = split /\s+/xms, $line;

      my $fields_db = {read_as           => $temp[0],
                       really_a          => $temp[1],
                       really_c          => $temp[2],
                       really_g          => $temp[3],## no critic (ProhibitMagicNumbers);
                       really_t          => $temp[4],## no critic (ProhibitMagicNumbers);
                       rescore           => $args_ref->{rescore},
                       id_run_tile       => $args_ref->{id_run_tile},
                      };
      $self->schema->resultset('ErrorsByNucleotide')->update_or_create($fields_db);

    }

    return 1;
  }

  sub _errors_by_cycle_and_nucleotide {
    my ($self, $file, $args_ref) = @_;

    my ($wanted) = $file =~ /Full[ ]breakdown[ ]of[ ]errors[ ]by[ ]cycle[ ]and[ ]nucleotide\n(.*?)\nError[ ]rate[ ]relative[ ]to[ ]reference[ ]by[ ]cycle\/nucleotide/xms;
    my @lines = split /\n/xms, $wanted;
    splice @lines, 0, 2;

    foreach my $line (@lines) {

      my @temp = split /\s+/xms, $line;
      my $fields_db = {cycle             => $temp[0],
                       read_as           => $temp[1],
                       really_a          => $temp[2],
                       really_c          => $temp[3],## no critic (ProhibitMagicNumbers);
                       really_g          => $temp[4],## no critic (ProhibitMagicNumbers);
                       really_t          => $temp[5],## no critic (ProhibitMagicNumbers);
                       rescore           => $args_ref->{rescore},
                       id_run_tile       => $args_ref->{id_run_tile},
                      };
      $self->schema->resultset('ErrorsByCycleAndNucleotide')->update_or_create($fields_db);
    }

    return 1;
  }

  sub _most_common_word {
    my ($self, $file, $args_ref) = @_;

    my ($wanted) = $file =~ /The[ ]\d+[ ]most[ ]common[ ]words[ ]with[ ]2[ ]blanks[ ]or[ ]less[:]\n(.*?)\n\nThe/xms;
    my @lines;
    if ($wanted) {
      @lines = split /\n/xms, $wanted;
    }
    shift@lines;

    foreach my $line (@lines) {

      my @temp = split /\s+/xms, $line;
      my $ranking = shift @temp;
      my $occurrence = shift @temp;

      my $fields_db = {rank           => $ranking,
                       occurrence     => $occurrence,
                       rescore           => $args_ref->{rescore},
                       id_run_tile       => $args_ref->{id_run_tile},
                      };

      my $number_of_sequences;
      if ($temp[0] =~ /\A\d+\z/xms) {

        $fields_db->{number_of_sequences} = $temp[0];
        $self->schema->resultset('MostCommonWord')->update_or_create($fields_db);
        next;
      }

      $number_of_sequences = scalar @temp;
      $fields_db->{number_of_sequences} = $number_of_sequences;
      foreach my $word (@temp) {
        $fields_db->{word} = $word;
        $self->schema->resultset('MostCommonWord')->update_or_create($fields_db);
      }
    }
    return 1;
  }

  sub _most_common_blank_pattern {
    my ($self, $file, $args_ref) = @_;

    my ($wanted) = $file =~ /[(]N=any[ ]nonblank[ ]character[)]\n(.*?)\n\nLog/xms;
    my @lines;
    if ($wanted) {
      @lines = split /\n/xms, $wanted;
    }
    shift@lines;

    foreach my $line (@lines) {

      my @temp = split /\s+/xms, $line;
      my $ranking = shift @temp;
      my $occurrence = shift @temp;

      my $fields_db = {rank           => $ranking,
                       occurrence     => $occurrence,
                       rescore           => $args_ref->{rescore},
                       id_run_tile       => $args_ref->{id_run_tile},
                      };

      my $number_of_sequences;
      if ($temp[0] =~ /\A\d+\z/xms) {
        $fields_db->{number_of_sequences} = $temp[0];
        $self->schema->resultset('MostCommonBlankPattern')->update_or_create($fields_db);
        next;
      }

      $number_of_sequences = scalar @temp;
      $fields_db->{number_of_sequences} = $number_of_sequences;

      foreach my $word (@temp) {
        $fields_db->{pattern} = $word;
        $self->schema->resultset('MostCommonBlankPattern')->update_or_create($fields_db);
      }

    }

    return 1;
  }
no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=item MooseX::Getopt

=item npg_qc::api::loader::base

=item Perl6::Slurp

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2009 Guoying Qi (gq1@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
