#############
# $Id: base.pm 13782 2011-08-01 14:48:20Z ajb $
# Created By: ajb
# Mast Maintained By: $Author: ajb $
# Created On: 2009-12-03
# Last Changed On: $Date: 2011-08-01 15:48:20 +0100 (Mon, 01 Aug 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-qc/tags/release-38.1/lib/npg_qc/api/loader/base.pm $

package npg_qc::api::loader::base;
use strict;
use warnings;
use Moose;
use Carp qw{carp cluck croak confess};
use English qw{-no_match_vars};
use Readonly;

use npg_qc::Schema;

with qw{MooseX::Getopt npg_common::roles::run::short_info npg_common::roles::run::path_info npg_common::roles::log};
with qw{npg_common::roles::run::long_info};
with qw(npg_common::roles::run::folder_validation);

# this is a general utility function for npg_qc::api object
use npg_qc::api::util;

Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 13782 $ =~ /(\d+)/mxs; $r; };
## no critic (Documentation::RequirePodAtEnd)

=head1 NAME

npg_qc::api::loader::base

=head1 VERSION

$Revision: 13782 $

=head1 SYNOPSIS

  my $oDerived = npg_qc::api::loader::<derived_class>->new({
    id_run => $iIdRun,
    verbose => 1,
    run_folder => $sRunFolder,
  });

id_run or run_folder must be provided, or at least a path for npg_common::roles::run::path_info
which should expose the run_folder. verbose turns on the optional logs

=head1 DESCRIPTION

This is intended to be a base class object for loader object modules. It exports various methods to the class,
along with consuming roles which are needed for all.

=head1 SUBROUTINES/METHODS

=head2 verbose

Boolean to be set on object creation

=cut

has q{verbose} => (isa => q{Bool}, is => q{ro},
  documentation => q{Some output only logged if verbose mode on},);

=head2 util

store on creation or retrieve a util object, ideally just a npg_qc::api::util, although if you choose something else,
you may find breakages, unless the methods match. This will lazy_build a default npg_qc::api::util object for you.

=cut

has q{util} => (isa => q{Object}, is => q{ro}, lazy_build => 1,
  documentation => q{Takes a util object, will default build a npg_qc::api::util object},);

sub _build_util {
  return npg_qc::api::util->new({
    useragent => undef,
  });
}

# private methods (not builders for public lazy_build attributes - excepting _build_run_folder)

# this class ties together short_info and path_info, so the following _build_run_folder will work
sub _build_run_folder {
  my ($self) = @_;
  my @temp = split m{/}xms, $self->runfolder_path();
  my $run_folder = pop @temp;
  return $run_folder;
}

=head2 schema

create an ojbect of dbix class schema

=cut

has q{schema} => (isa => q{Object}, is => q{ro}, lazy_build => 1,
  documentation => q{default build a npg_qc::Schema object},);

sub _build_schema {
  my $self = shift;
  return $self->util->schema();
}

=head2 run

the main method to start the whole process

=cut

sub run {
  my $self = shift;
  $self->npg_api_util($self->util->get_npg_api_util);
  if(!$self->check()){
     croak q{Run folder }. $self->runfolder_path(). q{ is not a genuine run in NPG.};
  }else{
     $self->log ('Runfolder '. $self->run_folder(). ' checked by NPG');
  }

  return 1;
}

=head2 get_id_run_tile

given lane, tile and read number, find or create id_run_tile in the database 

=cut

sub get_id_run_tile {
  my ($self, $lane_number, $read_number, $tile_number) = @_;

  my $id_run = $self->id_run();
  my $schema = $self->schema();

  my $run_tile = $schema->resultset('RunTile')
                        ->find_or_create(
                            {id_run    => $id_run,
                             position  => $lane_number,
                             tile      => $tile_number,
                             end       => $read_number,
                            });

  return $run_tile->id_run_tile();
}

=head2 get_id_analysis

given lane, tile and read number, find or create id_run_tile in the database 

=cut
sub get_id_analysis {
  my ($self, $end) = @_;

  my $id_run = $self->id_run();

  my ($y, $m, $d) = $self->run_folder() =~ /(\d\d)(\d\d)(\d\d)_\w{2}\d+_\d+/mxs;
  my $date = sprintf q(%4d-%02d-%02d), $y+2000, $m, $d;## no critic (Policy::ValuesAndExpressions::ProhibitMagicNumbers)

  my $folder = $self->recalibrated_path();
  if(!$folder){
    $folder = $self-> $self->bustard_path();
  }

  my $schema = $self->schema();

  my @current_analysis_per_run = $schema->resultset('Analysis')
                         ->search(
                            {id_run    => $id_run,
                             end       => $end,
                             iscurrent => 1,
                            });

  my @current_analysis_per_run_same_folder;

  foreach my $analysis (@current_analysis_per_run){
    if( $analysis->folder() eq $folder){
      push @current_analysis_per_run_same_folder, $analysis;
    } else {
      $analysis->update({iscurrent => 0});
    }
  }

  if(scalar @current_analysis_per_run_same_folder > 1){
    croak qq{More than one current rows found in analysis table for run $id_run, read $end, folder $folder};
  }elsif(scalar @current_analysis_per_run_same_folder == 1){
    return $current_analysis_per_run_same_folder[0]->id_analysis();
  }

  my $analysis = $schema->resultset('Analysis')
                        ->create(
                            {id_run    => $id_run,
                             end       => $end,
                             date      => $date,
                             folder    => $folder,
                             iscurrent => 1,
                            });

  return $analysis->id_analysis();
}

=head2 transfer_read_number

given read number for a multiplex run, transfer the read number 1->1, 2->t, 3->2
for non-indexing run, no change

=cut
sub transfer_read_number{
  my ($self, $read_number) = @_;

  if($read_number == 3){ ## no critic (Policy::ValuesAndExpressions::ProhibitMagicNumbers)
       $read_number = 2;
  }elsif($read_number == 2 && $self->is_indexed()){
       $read_number = q{t};
  }
  return $read_number;
}
no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=item MooseX::Getopt

=item npg_common::roles::run::short_info

=item npg_common::roles::run::path_info

=item npg_common::roles::log

=item npg_qc::Schema

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2009 Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
