#########
# Author:        Aylwyn Scally
# Maintainer:    $Author: ajb $
# Created:       2008-11-26
# Last Modified: $Date: 2011-08-01 15:48:20 +0100 (Mon, 01 Aug 2011) $
# Id:            $Id: gc_bias.pm 13782 2011-08-01 14:48:20Z ajb $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-qc/tags/release-38.1/lib/npg_qc/autoqc/checks/gc_bias.pm $
#
#

package npg_qc::autoqc::checks::gc_bias;

use strict;
use warnings;
use Moose;
use Carp;
use English qw(-no_match_vars);
use Fatal qw(open close);
use File::Basename;
use File::Slurp;
use IPC::Open3;
use List::Util qw(max);
use MIME::Base64::Perl;
use Perl6::Slurp;
use POSIX qw(WIFEXITED);
use Time::HiRes qw( gettimeofday tv_interval );

use npg_qc::autoqc::types;

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 13782 $ =~ /(\d+)/mxs; $r; };

extends qw(npg_qc::autoqc::checks::check);
with qw(npg_common::sequence::reference::roles::find);

Readonly::Scalar my $MINIMUM_WINDOW_SIZE =>     100;
Readonly::Scalar my $TARGET_WINDOW_COUNT => 150_000;
Readonly::Scalar my $R_EXECUTABLE        => '/software/bin/R';
Readonly::Scalar my $ALIGNER_REPOSITORY
    => '/software/solexa/bin/aligners/bin';

Readonly::Scalar our $EXT                => 'bam';


has '+input_file_ext' => (default    => $EXT,);

has 'samtools_path'  => ( is         => 'ro',
                          isa        => 'Str',
                          default    => $ALIGNER_REPOSITORY,
);

has 'r_executable'   => ( is         => 'ro',
                          isa        => 'Executable',
                          default    => $R_EXECUTABLE,
);

has 'samtools'       => ( is         => 'ro',
                          isa        => 'Executable',
                          lazy_build => 1,
);

has 'window_depth'   => ( is         => 'ro',
                          isa        => 'Executable',
                          lazy_build => 1,
);

has 'bam_file'       => ( is         => 'ro',
                          isa        => 'Str',
                          writer     => '_set_bam_file',
                          lazy_build => 1,
);

has 'gcn_file'       => ( is         => 'ro',
                          isa        => 'Str',
                          writer     => '_set_gcn_file',
                          lazy_build => 1,
);

has 'reference_base' => ( is         => 'ro',
                          isa        => 'Maybe[Str]',
                          writer     => '_set_reference_base',
);


has 'read_length'    => (is         => 'ro',
                         isa        => 'Maybe[Int]',
                         lazy_build => 1,
			);


has 'window_size'    => ( is         => 'ro',
                          isa        => 'Int',
                          writer     => '_set_window_size',
);


sub _build_samtools {
    my ($self) = @_;
    return $self->samtools_path() . q{/samtools};
}

sub _build_window_depth {
    my ($self) = @_;

    my $pid = open3( undef, my $out, undef, 'which window_depth' );
    waitpid $pid, 0;
    my $path = <$out>;
    defined $path or croak 'window_depth executable not found';
    chomp $path;

    return $path;
}

sub _build_bam_file {
    my ($self) = @_;
    return $self->input_files->[0];
}


sub _build_read_length {
    my ($self) = @_;

    $self->_set_ext(q[fastq]);
    my @files = $self->get_input_files();
    if (!@files) { croak q[No fastq files found]; }
    my $awk_cmd = qq{head -2 $files[0] | awk 'NR == 2 {print length}'};
    my $pid = open3( undef, my $out, undef, $awk_cmd );
    waitpid $pid, 0;
    my $read_length = <$out>;
    if ($read_length) {
        chomp $read_length;
    }
    return $read_length;
}


override 'can_run' => sub {
    my $self = shift;

    my $ref_arrayref;
    eval {
        $ref_arrayref = $self->refs();
        1;
        }
        or do {
        carp 'Could not eval refs() call';
        return 0;
        };

    my $reference = ( pop @{ $ref_arrayref } ) || q{};
    $self->_set_reference_base($reference);

    return ( -r $reference . q{.bwt} ? 1 : 0 );
};


override 'execute' => sub {
    my ($self) = @_;

    return 1 if super() == 0;

    if (!defined $self->read_length) {
        $self->result->add_comment('read length not found');
        return 1;
    }
    if ($self->read_length <= 0) {
        $self->result->add_comment('read length negative or zero: ' . $self->read_length);
        return 1;
    }

    if ( !$self->can_run() ) {
        $self->result->add_comment('No reference. Cannot run gc bias check.');
        return 1;
    }

    my $reference_fasta = $self->reference_base();
    $reference_fasta =~ s{/bwa/}{/fasta/}msx;
    if ( !-r $reference_fasta ) {
        $self->result->add_comment('No reference fasta. Cannot run gc bias check.');
        return 1;
    }

    my $reference_size = -s $reference_fasta;
    my $window_size = $self->select_window_size($reference_size);
    $self->_set_window_size($window_size);
    $self->result->window_size($window_size);

    my $reference_gcn = $self->reference_base();
    $reference_gcn =~ s{/bwa/}{/npgqc/}msx;
    $reference_gcn .= q{.gcn} . $self->window_size();

    if ( !-r $reference_gcn ) {
        $self->result->add_comment(qq[No gcn file $reference_gcn. Cannot run gc bias check]);
        return 1;
    }

    $self->_set_gcn_file($reference_gcn);

    my $tempdir         = $self->tmp_path();
    my ($bam_name, $xpath, $xsuffix) = fileparse($self->bam_file);
    my $output_basename = $tempdir . q{/} . $bam_name;
    my $plot_png        = $output_basename . q{-gc_bias.png};
    my $parameter_file  = $output_basename . q{-gcdepth.txt};
    my $depth_out       = $output_basename . q{.depth};

    my $depth_command = $self->samtools() . q{ view } . $self->bam_file()
                      .  q{ | } . $self->window_depth()
                      . qq{ $reference_fasta -b=} . $self->window_size()
                      .  q{ -g=} . $self->gcn_file()
                      . qq{ > $depth_out};

    croak 'window depth failed' if !WIFEXITED( system $depth_command );

    $self->create_gc_bias_graph($depth_out, $tempdir, $bam_name );

    # Because of building this module around inherited code we're using R to
    # calculate the various parameters for the plot, which are then being
    # written to a text file. We now have to read them from that text file.
    # This is a bit crap, and we should try to have perl calculate the
    # parameters itself.

    if ( !-r $parameter_file ) {
        $self->result->add_comment('R script failed. Data output not found.');
        return 1;
    }
    if ( !-r $plot_png ) {
        $self->result->add_comment('R script failed. Plot output not found.');
        return 1;
    }

    my $r_output = slurp $parameter_file;
    my $png_data = read_file( $plot_png, binmode => ':raw' );

    my ($max_y)        = $r_output =~ m/ DEPMAX:      \s+(.*?)\n /msx;
    my ($window_count) = $r_output =~ m/ WINDOW_COUNT:\s+(.*?)\n /msx;
    my ($bin_count)    = $r_output =~ m/ BIN_COUNT:   \s+(.*?)\n /msx;
    my ($polygon_x)    = $r_output =~ m/ POLYGON_X:   \s+(.*?)\n /msx;
    my ($polygon_y)    = $r_output =~ m/ POLYGON_Y:   \s+(.*?)\n /msx;
    my ($gc_tickmark)  = $r_output =~ m/ GC_TICKMARK: \s+(.*?)\n /msx;
    my ($main_plot_x)  = $r_output =~ m/ MAIN_PLOT_X: \s+(.*?)\n /msx;
    my ($main_plot_y)  = $r_output =~ m/ MAIN_PLOT_Y: \s+(.*?)\n /msx;
    my ($lower_poiss)  = $r_output =~ m/ LOWER_POISS: \s+(.*?)\n /msx;
    my ($upper_poiss)  = $r_output =~ m/ UPPER_POISS: \s+(.*?)\n /msx;

    $self->result->max_y(        $max_y        + 0 );
    $self->result->window_count( $window_count + 0 );
    $self->result->bin_count(    $bin_count    + 0 );
    $self->result->actual_quantile_x(    [ split m/,/msx, $polygon_x ] );
    $self->result->actual_quantile_y(    [ split m/,/msx, $polygon_y ] );
    $self->result->gc_lines(             [ split m/,/msx, $gc_tickmark ] );
    $self->result->plot_x(               [ split m/,/msx, $main_plot_x ] );
    $self->result->plot_y(               [ split m/,/msx, $main_plot_y ] );
    $self->result->ideal_lower_quantile( [ split m/,/msx, $lower_poiss ] );
    $self->result->ideal_upper_quantile( [ split m/,/msx, $upper_poiss ] );
    $self->result->cached_plot( encode_base64($png_data) );

    return 1;
};


sub select_window_size {
    my ( $self, $reference_size ) = @_;

    my $window_size = max( $self->read_length,
                           $MINIMUM_WINDOW_SIZE,
                           $reference_size / $TARGET_WINDOW_COUNT );
    $window_size = up_nearest($window_size);

    return $window_size;
}


sub up_nearest {
    my ($number) = @_;
    $number = 0 + $number;
    return 0 if !$number;

    my $factor = $number <=> 0;
    $number = abs $number;

    ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)
    while ( $number >= 10 ) { $number /= 10; $factor *= 10; }
    while ( $number <   1 ) { $number *= 10; $factor /= 10; }

    ## use critic
    return ( int($number) == $number ? $number : ( int($number) + 1 ) )
           * $factor;
}


sub create_gc_bias_graph {
    my ( $self, $bindepth_file, $output_path, $plot_name) = @_;

    local $ENV{R_PROFILE_USER} = $ENV{HOME} . q{/.Rprofile};

    my $r_file      = $bindepth_file . q{.R};
    $output_path = $output_path . q[/];
    my $r_lib = $self->find_R_library();

    # Create the R script and run the command.
    open my $fh, q{>}, $r_file;

    print {$fh} qq[source('$r_lib')\n],
                qq[depdat = read.depth('$bindepth_file')\n],
                qq[gcdepth(depdat, plot_name = '$plot_name', depmax = NULL,],
                 q[        nbins = 30, plotdev = bitmap, binned = TRUE,],
                qq[        path = '$output_path', make_plot = TRUE)\n]
        || croak $OS_ERROR;

    close $fh;

    my $plot_command = qq{cat $r_file | }
                     . $self->r_executable()
                     . q{ --slave --no-save --no-site-file --no-restore};

    croak "Calculate plot data failed: $OS_ERROR"
        if !WIFEXITED( system $plot_command );

    unlink $r_file;

    return;
}


sub find_R_library {    ## no critic (NamingConventions::Capitalization)
    my ($self) = @_;

    my @path_bits = split m{/}msx, dirname(__FILE__);

    while ( scalar @path_bits > 0 ) {
        last if $path_bits[-1] eq 'lib';
        pop @path_bits;
    }

    if ( scalar @path_bits == 0 ) {
        $self->result->
            add_comment('Parent \'lib/R\' directory not found for R script');
    }

    my $path_to_R_lib = join q{/}, @path_bits; ## no critic (NamingConventions::Capitalization)
    $path_to_R_lib .= q{/R/gc_bias_data.R};

    return $path_to_R_lib;
}


no Moose;
__PACKAGE__->meta->make_immutable();


1;

__END__


=head1 NAME

npg_qc::autoqc::checks::gc_bias - assess the degree of gc_bias in reads
    aligned to the reference.

=head1 VERSION

    $Revision: 13782 $

=head1 SYNOPSIS

    use npg_qc::autoqc::checks::gc_bias();


=head1 DESCRIPTION


=head1 SUBROUTINES/METHODS

=head2 new

    Moose-based.

=head2 up_nearest

    Round a number up to the nearest, single significant, digit integer. E.g.
    100, 7000, 300000, etc.

=head2 select_window_size

    Select the window size to use based on reference size and read length.

=head2 create_gc_bias_graph

    Call the R code to generate the numbers for the plot, and to create the
    png version of the plot.

=head2 find_R_library

    To avoid additional configuration steps (usually creating ~/.Rprofile) the
    R library is included with the perl modules. Find it by traversing up the
    path to this module until the directory 'lib' is found, then descend to
    'R/gc_bias_data.R'.

=head1 DIAGNOSTICS

    None.

=head1 CONFIGURATION AND ENVIRONMENT

=head1 INCOMPATIBILITIES

    None known.

=head1 BUGS AND LIMITATIONS

=head1 DEPENDENCIES

=head1 AUTHOR

    John O'Brien, jo3

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by John O'Brien

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
