#########
# Author:        Andy Brown ajb@sanger.ac.uk and Marina Gourtovaia mg8@sanger.ac.uk 
# Maintainer:    $Author: ajb $
# Created:       Summer 2009
# Last Modified: $Date: 2011-08-01 15:48:20 +0100 (Mon, 01 Aug 2011) $
# Id:            $Id: qc_store.pm 13782 2011-08-01 14:48:20Z ajb $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-qc/tags/release-38.1/lib/npg_qc/autoqc/qc_store.pm $
#

package npg_qc::autoqc::qc_store;
use strict;
use warnings;
use English qw(-no_match_vars);;
use Readonly;
use Moose;
use Carp;

use npg_common::config;

use npg_qc::autoqc::db::Schema;
use npg_qc::autoqc::db::DB;
use npg_qc::autoqc::results::collection;

our $VERSION   = do { my ($r) = q$Revision: 13782 $ =~ /(\d+)/mxs; $r; };
## no critic (Documentation::RequirePodAtEnd)

=head1 NAME

npg_qc::autoqc::qc_store

=head1 VERSION

$Revision: 13782 $

=head1 SYNOPSIS

=head1 DESCRIPTION

Transparent retrieval of stored auto QC data.
One instance of this class should be created by the application.
If used as a model in the catalyst application this class is
likely to be instantiated at server start up time

=head1 SUBROUTINES/METHODS

=cut


Readonly::Scalar  our $DO_USE_DB => q[yes];
Readonly::Scalar  our $NPG_CONF_DIR => q[.npg];
Readonly::Scalar  our $MIN_STAGING_ID_RUN => 5000;


=head2 use_db

A flag indicating whether a DB lookup will be performed.
Possible values 'yes' and 'no'.
If the flag is not set to 'yes', data will be retrieved from
the staging area only; no attempt to connect to the QC database
will be made.

=cut
has 'use_db'       => ( isa        => 'Str',
                        is         => 'ro',
                        required   => 0,
                        default    => $DO_USE_DB,
		      );

=head2 recent_staging_runs

A flag indicating whether a lookup on a filesystem should be
performed only for recent runs. Defaults to false.

=cut
has 'recent_staging_runs'       => ( isa        => 'Bool',
                                     is         => 'ro',
                                     required   => 0,
                                     default    => 0,
		                   );

=head2 min_recent_staging_run

Minimun id_run that is considered recent.

=cut
has 'min_recent_staging_run'    => ( isa        => 'Int',
                                     is         => 'ro',
                                     required   => 0,
                                     default    => $MIN_STAGING_ID_RUN,
		                   );

=head2 dsn

A dsn string for connecting to the QC database

=cut
has 'dsn'          => ( isa        => 'Str',
                        is         => 'ro',
                        required   => 0,
                      );

=head2 user

The user name for connecting to the QC database

=cut
has 'user'         => ( isa        => 'Str',
                        is         => 'ro',
                        required   => 0,
                      );

=head2 password

The password for connecting to the QC database

=cut
has 'password' =>     ( isa        => 'Str',
                        is         => 'ro',
                        required   => 0,
                      );


has '_db_lookup' =>   ( isa        => 'Bool',
                        is         => 'ro',
                        required   => 0,
                        default    => 1,
                        writer     => '_write_db_lookup',
                      );


=head2 BUILD

A post-constructor step, called before returning a ref to a new object to the caller.
Connects to the QC database if necessary.

=cut
sub BUILD {

    my $self = shift;
    if ($self->use_db eq $DO_USE_DB) {

        my $dsn = $self->dsn || q[];
        my $user = $self->user || q[];
        my $password = $self->password || q[];

        if (!$dsn) {
            my $config = npg_common::config->new(
                                conf_dir  => $NPG_CONF_DIR,
                                conf_name => q[npg_qc-Schema],
                                                );
            $dsn = $config->dsn;
            $user = $config->dbuser;
            $password = $config->dbpass;

	}

        if (!$dsn) {
            croak (q[Cannot connect to the qc DB without a dsn string]);
        }
        npg_qc::autoqc::db::Schema->load_schema({
                             dsn => $dsn,
                             username => $user,
                             password => $password
                                               });
    } else {
        $self->_write_db_lookup(0);
    }
    return 1;
}


=head2 load_run

Data for a full run returned as an npg_qc::autoqc::results::collection object.

=cut
sub load_run {
    my ($self, $id_run, $db_lookup, $lanes, $what) = @_;

    if (!$id_run) { croak q[id_run should be set]; }

    if (!defined $db_lookup) { $db_lookup = 1; }

    my  $collection = undef;
    if ($self->_db_lookup && $db_lookup) {
        $collection = npg_qc::autoqc::db::DB->autoqc_data_by_run_db($id_run, $lanes, $what);
    }
    if ( !defined $collection || $collection->is_empty() ) {
        $collection = npg_qc::autoqc::results::collection->new();
        if (!$self->recent_staging_runs || $id_run > $self->min_recent_staging_run) {
            ## no critic (RequireCheckingReturnValueOfEval)
            eval {
               $collection->add_from_staging($id_run, $lanes, $what);
            };
        }
    }

    return $collection;
}


=head2 load_lanes

Data for a number of lanes returned as an npg_qc::autoqc::results::collection object.
The second argument is a reference to a hash defining the lanes. Run Ids should be
used as keys. A reference to an array with lane numbers should be used as values.

=cut
sub load_lanes {
    my ($self, $run_lanes_hash, $db_lookup, $what) = @_;

    my $collection =  npg_qc::autoqc::results::collection->new();
    foreach my $id_run (keys %{$run_lanes_hash}) {
        my $temp_collection = $self->load_run($id_run, $db_lookup, $run_lanes_hash->{$id_run}, $what);
        if ($temp_collection && $temp_collection->size > 0) {
            $collection->add($temp_collection->results);
        }
    }
    return $collection;
}


=head2 load_from_path

Reads all serialised result objects from a given path or a list of paths and returns them as 
an npg_qc::autoqc::results::collection object. The second argument is a reference
to an run id in the code of the caller. An id_run value is assigned to it.

=cut
sub load_from_path {
  my ($self, @path) = @_;
  if (!@path) {
    croak 'Path should be given';
  }
  my $c = npg_qc::autoqc::results::collection->new();
  foreach my $p (@path) {
    $c->add_from_dir($p);
  }
  return $c;
}

no Moose;
__PACKAGE__->meta->make_immutable;

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English

=item Readonly

=item strict

=item warnings

=item npg_common::config

=item npg_qc::autoqc::autoqc

=item npg_qc::autoqc::db::Schema

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Marina Gourtovaia E<lt>mg8@sanger.ac.ukE<gt> and Andy Brown E<lt>ajb@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Marina Gourtovaia and Andy Brown

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut

