#########
# Author:        jo3
# Maintainer:    $Author: ajb $
# Created:       Autumn 2010
# Last Modified: $Date: 2011-08-01 15:48:20 +0100 (Mon, 01 Aug 2011) $
# Id:            $Id: 60-autoqc-checks-ref_match.t 13782 2011-08-01 14:48:20Z ajb $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-qc/tags/release-38.1/t/60-autoqc-checks-ref_match.t $
#
use strict;
use warnings;
use English qw(-no_match_vars);
use autodie qw(:all);
use File::Temp qw/ tempdir /;
use Cwd qw/cwd/;

use Test::More tests => 45;
use Test::Deep;
use Test::Exception;

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 13782 $ =~ /(\d+)/msx; $r; };


use_ok('npg_qc::autoqc::checks::ref_match');

my $fastq_path = 't/data/autoqc';
my $repos = join q[/], cwd, q[t/data/autoqc];
my $ref_repos  = 't/data/autoqc/references';

{
    my $test;

    lives_ok {
        $test = npg_qc::autoqc::checks::ref_match->new(
                        path     => $fastq_path,
                        id_run   => 1937,
                        position =>    3,
                        repository => $repos,
        );
    }
    'Create the minimal check object';

    isa_ok( $test, 'npg_qc::autoqc::checks::ref_match' );

    # Test that default options are set. But don't worry about the actual
    # values. Changes are less hassle that way, plus it improves
    # Sanger-independence. We'll test values later anyway.
    ok( defined $test->aligner(), 'A default aligner is set' );

    ok( defined $test->aligner_path(),
        'A default path to the aligner is set' );

    ok( defined $test->aligner_options(), 'Some default options are set' );

    ok( defined $test->ref_repository(),
        'A default reference repository is set' );

    my $temp_dir = $test->tmp_path();
    ok( -d $temp_dir, 'Temporary directory created' );

    ok( $test->temp_fastq() =~ m{^ $temp_dir / \S+ }msx,
        'Temporary fastq path is in the temp dir' );

    ok( $test->sample_read_length() =~ m/^ \d+ $/msx,
        'A default read length is set' );
    ok( $test->sample_read_count()  =~ m/^ \d+ $/msx,
        'A default read count is set' );
}


{
    my $override_test;

    lives_ok {
        $override_test = npg_qc::autoqc::checks::ref_match->new(
                        path               => $fastq_path,
                        id_run             => 1937,
                        position           =>    3,
                        repository         => $repos,
                        aligner            => 'use_this_one',
                        aligner_path       => '/aligner/collection',
                        aligner_options    => '--switches go --here',
                        temp_fastq         => '/tEmp/faStq',
                        sample_read_length => 250,
                        sample_read_count  => 3,
                        request_list       => [ 'abc', 'def' ],
                        ref_repository     => 't/data',
        );
    }
    'Various things can be set if needed';

    isa_ok( $override_test, 'npg_qc::autoqc::checks::ref_match' );

    is( $override_test->aligner(), 'use_this_one',
        'Default aligner overridden' );
    is( $override_test->aligner_path(), '/aligner/collection',
        'Default aligner path overridden' );
    is( $override_test->aligner_options(), '--switches go --here',
        'Default aligner options overridden' );
    is( $override_test->temp_fastq(), '/tEmp/faStq',
        'Default temporary fastq path overridden' );
    is( $override_test->sample_read_length(), 250,
        'Default read length overridden' );
    is( $override_test->sample_read_count(), 3,
        'Default read count overridden' );
    cmp_bag( $override_test->request_list(), [ 'abc', 'def' ],
        'Requested organism list explicitly specified' );
    is( $override_test->ref_repository(), 't/data',
        'Default reference repository overridden' );
}

{
    throws_ok {
        npg_qc::autoqc::checks::ref_match->new(
                        path               => $fastq_path,
                        id_run             => 1937,
                        position           =>    3,
                        sample_read_length =>   27,
                        sample_read_count  =>    3,
                        repository => $repos,
        )
    } qr/Sample read length 27 is below 28 \(lowest acceptable value\)/,
    'error when requesting too short reads';
}

{
    my $test = npg_qc::autoqc::checks::ref_match->new(
                        path               => $fastq_path,
                        id_run             => 1937,
                        position           =>    3,
                        sample_read_length =>   37,
                        sample_read_count  =>    3,
                        read1_fastq        => q[t/data/autoqc/ref_match/narrow.fastq],
                        repository => $repos,
    );
    lives_ok { $test->execute() }
             'No error when reads are shorter than requested';
    is($test->result->comments, q[Read length of 4 is below minimally required 28], 'comment for a v short read');
}


{
    my $test;
    lives_ok {
        $test = npg_qc::autoqc::checks::ref_match->new(
                        path               => $fastq_path,
                        id_run             => 1937,
                        position           =>    3,
                        sample_read_length =>   39,
                        sample_read_count  =>    3,
                        repository => $repos,
        );
    }
    'Set sample length to 38 for a 37 bp read length';

    lives_ok { $test->_create_sample_fastq() }
             'No error when reads are shorter than requested';
    ok( -e $test->temp_fastq(), 'Temp fastq file created' );
    is( $test->result->sample_read_length(), 37, 'Sample read length is 37' );
    is( $test->result->sample_read_count(),   3, 'Sample read count is 3' );
}


{
    my $test = npg_qc::autoqc::checks::ref_match->new(
                        path               => $fastq_path,
                        id_run             => 1937,
                        position           =>    2,
                        repository => $repos,
    );

    ok(!$test->_create_sample_fastq(), 'creating sample fastq for an empty file returns false');
    lives_ok {$test->execute()} 'execute method for an empty fastq file lives';
    like( $test->result->comments(), qr/[.]fastq[ ]is[ ]empty/msx,
          'Store an appropriate comment for an empty fastq file' );
}


{
    my $test = npg_qc::autoqc::checks::ref_match->new(
                path           => $fastq_path,
                id_run         => 1937,
                position       =>    3,
                ref_repository => $ref_repos,
                repository => $repos,
    );

    isa_ok( $test, 'npg_qc::autoqc::checks::ref_match', 'The test object for testing organism list building' );

    cmp_bag( $test->organism_list(), [ 'Homo_sapiens', 'Vibrio_cholerae' ],
             'Default organism list is correct' );

    is_deeply( $test->reference_version(),
               {
                'Homo_sapiens'    => 'NCBI36',
                'Vibrio_cholerae' => 'M66-2'
               },
               'Default reference version list is correct'
    );

    is_deeply( $test->index_base(),
               {
                'Homo_sapiens' => $ref_repos .
                            '/Homo_sapiens/default/all/bowtie/someref.fa',
                'Vibrio_cholerae' => $ref_repos .
                            '/Vibrio_cholerae/default/all/bowtie/V_chol.fasta'
               },
               'Default index base list is correct'
    );


    $test = npg_qc::autoqc::checks::ref_match->new(
                path           => $fastq_path,
                id_run         => 1937,
                position       =>    3,
                ref_repository => $ref_repos,
                request_list   => ['Vibrio_cholerae'],
                repository => $repos,
    );

    cmp_bag( $test->organism_list(), ['Vibrio_cholerae'],
             'Request list argument effective' );
}


{
    my $test = npg_qc::autoqc::checks::ref_match->new(
                path            => $fastq_path,
                id_run          => 1937,
                position        =>    3,
                ref_repository  => $ref_repos,
                request_list    => [ 'Vibrio_cholerae' ],
                aligner_path    => '/bin/true',
                aligner_options => 'ref=%ref% reads=%reads%',
                repository => $repos,
    );

    my $comm;
    lives_ok { $comm = $test->_align_command('Vibrio_cholerae'); } 'Call align method';

    my $expect = q{/bin/true}
               . q{ ref=}   . $ref_repos
               . q{/Vibrio_cholerae/default/all/bowtie/V_chol.fasta}
               . q{ reads=} . $test->temp_fastq();

    is( $comm, $expect, 'Aligner command is correctly constructed' );
}


{
    my $test = npg_qc::autoqc::checks::ref_match->new(
                path           => $fastq_path,
                id_run         => 1937,
                position       =>    3,
                ref_repository => $ref_repos,
                request_list   => ['Vibrio_cholerae'],
                repository => $repos,
    );

    my $sam =  't/data/autoqc/alignment_refmatch.sam';
    open my $fh , q[<], $sam;

    my $num_matches;
    lives_ok { $num_matches = $test->_parse_output($fh) } 'Call parse output method';
    is ($num_matches, 2, 'correct number of matches returned');
}

{
    my $dir = 	q[does_not_exist];
    my $bt = join q[/], $dir, q[bowtie];

    my $test = npg_qc::autoqc::checks::ref_match->new(
                path           => $fastq_path,
                aligner_path   => $bt,
                id_run         => 1937,
                position       =>    3,
                ref_repository => $ref_repos,
                request_list   => ['Vibrio_cholerae'],
                repository => $repos,
    );

    local $ENV{PATH} = join q[:], $dir, $ENV{PATH};
    throws_ok {$test->execute} qr/Cannot fork \"does_not_exist\/bowtie/, 'error when cannot start an aligner';
}

{
    my $dir = tempdir(CLEANUP => 1);
    my $bt = join q[/], $dir, q[bowtie];

    my $test = npg_qc::autoqc::checks::ref_match->new(
                path           => $fastq_path,
                aligner_path   => $bt,
                id_run         => 1937,
                position       =>    3,
                ref_repository => $ref_repos,
                request_list   => ['Vibrio_cholerae'],
                repository => $repos,
    );

    open my $fh,  q[>], $bt;
    print $fh qq[ls dodo\n];
    close $fh;
    `chmod 755 $bt`;
    local $ENV{PATH} = join q[:], $dir, $ENV{PATH};

    throws_ok {$test->execute} qr/(exited with status 2)|(Cannot close bad pipe)/, 'error when aligner dies';
}

{
    my $dir = tempdir(CLEANUP => 1);
    my $bt = join q[/], $dir, q[bowtie];

    my $test = npg_qc::autoqc::checks::ref_match->new(
                path           => $fastq_path,
                aligner_path   => $bt,
                id_run         => 1937,
                position       =>    3,
                ref_repository => $ref_repos,
                request_list   => ['Vibrio_cholerae'],
                repository => $repos,
    );

    open my $fh,  q[>], $bt;
    print $fh qq[cat t/data/autoqc/alignment_refmatch.sam\n];
    close $fh;
    `chmod 755 $bt`;
    local $ENV{PATH} = join q[:], $dir, $ENV{PATH};

    lives_ok {$test->execute} 'execute lives';
    is_deeply( $test->result->aligned_read_count(), { 'Vibrio_cholerae' => 2 }, 'Correct match count saved' );
}

1;
