use strict;
use warnings;
use Test::More tests => 45;
use Test::Exception;
use SangerPaths qw(core ensembl);
use File::Temp qw/tempdir/;

use t::api::util;
use npg_qc::autoqc::qc_store::options qw/$ALL $LANES $PLEXES/;
use npg_qc::autoqc::db::DB;


{ use_ok 'npg_qc::autoqc::qc_store' }

{
  lives_ok { npg_qc::autoqc::qc_store->new({use_db    => q[no],}) } 'no-db-credentials constructor - no error';
  lives_ok { npg_qc::autoqc::qc_store ->new({    
                                      use_db    => q[no],
                                      dsn       => q[dsn],
                                      username  => q[admin],
                                      password  => q[pword],
                                              }) 
           } 'with full args constructor - no error';
}

{
  isa_ok(npg_qc::autoqc::qc_store->new({use_db    => q[no],}), 'npg_qc::autoqc::qc_store');

  local $ENV{dev} = q[non-existing];
  throws_ok { npg_qc::autoqc::qc_store->new(use_db    => q[yes]) } qr/No config file found/, 'not found config file section error';
}

{
  my $model =  npg_qc::autoqc::qc_store->new(    
                                      use_db    => q[no],
                                      dsn       => q[dsn],
                                      username  => q[admin],
                                      password  => q[pword]
                                             );
  ok(!$model->_db_lookup, 'db_lookup set to false');
}

{
  my $model =  npg_qc::autoqc::qc_store->new(    
                                      use_db    => q[no],
                                      dsn       => q[dsn],
                                      username  => q[admin],
                                      password  => q[pword]
                                             );
  throws_ok {$model->load_run()} qr/id_run should be set/, 'error loading run when id_run is not set';
}

{
  my $model =  npg_qc::autoqc::qc_store->new(    
                                      use_db    => q[no],
                                      dsn       => q[dsn],
                                      username  => q[admin],
                                      password  => q[pword]
                                             );
  
  local $ENV{TEST_DIR} = q[t/data];
  my $collection = $model->load_run(1234);
  is($collection->size(), 16, 'number of qc results from staging when the db is not used');
}

{
  my $model =  npg_qc::autoqc::qc_store->new(    
                                      use_db    => q[no],
                                      dsn       => q[dsn],
                                      username  => q[admin],
                                      password  => q[pword],
                                      recent_staging_runs => 1
                                             );
  
  local $ENV{TEST_DIR} = q[t/data];
  my $collection = $model->load_run(1234);
  is($collection->size(), 0, 'no qc results from staging when the db is not used, run is old and recent_staging_runs is true');
}

{
  my $model =  npg_qc::autoqc::qc_store->new(    
                                      use_db    => q[no],
                                      dsn       => q[dsn],
                                      username  => q[admin],
                                      password  => q[pword],
                                      recent_staging_runs => 1,
                                      min_recent_staging_run => 1000
                                             );
  
  local $ENV{TEST_DIR} = q[t/data];
  my $collection = $model->load_run(1234);
  is($collection->size(), 16, 'number qc results from staging when the db is not used, run is recent and recent_staging_runs is true');
}

{
  local $ENV{dev} = 'test';

  my $util = t::api::util->new({fixtures => 1, dbh => undef, useragent => undef});
  my $dbname   = $util->dbname();
  my $dbhost   = $util->dbhost();
  my $dbport   = $util->dbport();
  my $user     = $util->dbuser();
  my $password = $util->dbpass();

 
  my $model;
  lives_ok {
    $model = npg_qc::autoqc::qc_store->new(    
          dsn       => "DBI:mysql:database=$dbname;host=$dbhost;port=$dbport",
          user      => $user,
          password  => $password,  );
  } 'no error connecting to the db';

  ok($model->_db_lookup, 'db_lookup set to true');

  local $ENV{TEST_DIR} = q[t/data];

  my $collection = $model->load_run(1234);
  is($collection->size(), 16, 'number of qc results from staging when the db is used');
}

{
  local $ENV{TEST_DIR} = q[t];
  my $model =  npg_qc::autoqc::qc_store->new(    
                                      use_db    => q[no],
                                      dsn       => q[dsn],
                                      username  => q[admin],
                                      password  => q[pword]
                                             );
  my $collection = $model->load_run(3500);
  is( $collection->size(), 0,
    'no qc_data from database for run 3500 with db flag unset' );  
}

{
  local $ENV{dev} = 'test';
  local $ENV{TEST_DIR} = q[t];
  my $util = t::api::util->new({fixtures => 1, dbh => undef, useragent => undef});
  my $dbname   = $util->dbname();
  my $dbhost   = $util->dbhost();
  my $dbport   = $util->dbport();
  my $user     = $util->dbuser();
  my $password = $util->dbpass();

 
  my $model;
  lives_ok {$model =  npg_qc::autoqc::qc_store->new(    
                                      dsn       => "DBI:mysql:database=$dbname;host=$dbhost;port=$dbport",
                                      user      => $user,
                                      password  => $password,
                                                    );
           } 'no error connecting to the db';

  ok($model->_db_lookup, 'db_lookup set to true');

  my $collection = $model->load_run(3500, 0);
  is( $collection->size(), 0,
    'no qc_data from database for run 3500 with db flag set but overridden in the method' );  
}

{
  local $ENV{dev} = 'test';
  local $ENV{TEST_DIR} = q[t/data];

  my $util = t::api::util->new({fixtures => 1, dbh => undef, useragent => undef});
  my $dbname   = $util->dbname();
  my $dbhost   = $util->dbhost();
  my $dbport   = $util->dbport();
  my $user     = $util->dbuser();
  my $password = $util->dbpass();

 
  my $model;
  lives_ok {$model =  npg_qc::autoqc::qc_store->new(    
                                      use_db    => q[yes],
                                      dsn       => "DBI:mysql:database=$dbname;host=$dbhost;port=$dbport",
                                      user      => $user,
                                      password  => $password,
                                                    );
           } 'no error connecting to the db';

  npg_qc::autoqc::db::DB->save_from_jsonfile("t/data/autoqc/3510_4.qx_yield.json");
  npg_qc::autoqc::db::DB->save_from_jsonfile("t/data/autoqc/3510_1.insert_size.json");
  npg_qc::autoqc::db::DB->save_from_jsonfile("t/data/autoqc/load/3510_5.qX_yield.json"); 
  
  my $db_lookup = 1;
  
  my $run_lanes = {3500 => [1],};
  is($model->load_lanes($run_lanes, $db_lookup)->size(), 8, 'db retrieval for run 3500 lane 1');
  
  $run_lanes = {3500 => [2,4],};
  is($model->load_lanes($run_lanes, $db_lookup)->size(), 0, 'db retrieval for run 3500 lane 2,4 (none exists)');

  $run_lanes = {3500 => [1,2],};
  is($model->load_lanes($run_lanes, $db_lookup)->size(), 8, 'db retrieval for run 3500 lane 1,2 (only 1 exists)');

  $run_lanes = {3510 => [1,5,4],};
  is($model->load_lanes($run_lanes, $db_lookup)->size(), 3, 'db retrieval for run 3510 lane 1,4,5 (all exist)');

  $run_lanes = {3500 => [1,2], 3510 => [1,5,4,8] };
  is($model->load_lanes($run_lanes, $db_lookup)->size(), 11, 'db retrieval for 2 runs, some positions do not exist');

  $run_lanes = {3500 => [1,2], 3510 => [1,5,4,8], 1234 => [8] };
  is($model->load_lanes($run_lanes, $db_lookup)->size(), 13, 'db+staging retrieval for 3 runs, some positions do not exist');
}

{
  local $ENV{dev} = 'test';

  my $util = t::api::util->new({fixtures => 1, dbh => undef, useragent => undef});
  my $dbname   = $util->dbname();
  my $dbhost   = $util->dbhost();
  my $dbport   = $util->dbport();
  my $user     = $util->dbuser();
  my $password = $util->dbpass();

 
  my $model;
  lives_ok {$model =  npg_qc::autoqc::qc_store->new(    
                                      dsn       => "DBI:mysql:database=$dbname;host=$dbhost;port=$dbport",
                                      user      => $user,
                                      password  => $password,
                                                    );
           } 'no error connecting to the db';

  ok($model->_db_lookup, 'db_lookup set to true');

  my $collection = $model->load_from_path('t/data/autoqc/rendered/json_paired_run');
  is( $collection->size(), 16, 'loading 16 results from path' );
  $collection = $model->load_from_path('t/data/autoqc/rendered/json_paired_run', 't/data/autoqc/rendered/json_paired_run');
  is( $collection->size(), 32, 'loading 16 results from path' );
  $collection = $model->load_from_path(('t/data/autoqc/rendered/json_paired_run', 't/data/autoqc/rendered/json_paired_run'));
  is( $collection->size(), 32, 'loading 16 results from path' ); 
}

{
  local $ENV{dev} = 'test';

  my $util = t::api::util->new({fixtures => 1, dbh => undef, useragent => undef});
  my $dbname   = $util->dbname();
  my $dbhost   = $util->dbhost();
  my $dbport   = $util->dbport();
  my $user     = $util->dbuser();
  my $password = $util->dbpass();

  my $model;
  lives_ok {$model =  npg_qc::autoqc::qc_store->new(    
                                      dsn       => "DBI:mysql:database=$dbname;host=$dbhost;port=$dbport",
                                      user      => $user,
                                      password  => $password,
                                                    );
           } 'no error connecting to the db';

  ok($model->_db_lookup, 'db_lookup set to true');

  my $id_run = 0;
  my $collection = $model->load_run(3600);
  is( $collection->size(), 4, '4 lane results from the db' ); 
  $collection = $model->load_run(3600, 1, []);
  is( $collection->size(), 4, '4 lane results from the db' );
  $collection = $model->load_run(3600, 1, [], $LANES);
  is( $collection->size(), 4, '4 lane results from the db' );
  $collection = $model->load_run(3600, 1, [], $ALL);
  is( $collection->size(), 10, '10 lane results from the db' );
  $collection = $model->load_run(3600, 1, [], $PLEXES);
  is( $collection->size(), 6, '6 plex results from the db' );
  $collection = $model->load_run(3600, 1, [4], $PLEXES);
  is( $collection->size(), 3, '3 plex lane 4 results from the db' );
  $collection = $model->load_run(3600, 1, [4], $ALL);
  is( $collection->size(), 4, '4 results from the db' );
}

{
  my $temp = tempdir( CLEANUP => 1);
  my $other =  join(q[/], $temp, q[nfs]);
  mkdir $other;
  $other = join(q[/], $other, q[sf44]);
  mkdir $other;

  `cp -R t/data/nfs/sf44/IL2  $other`;
  my $archive = join q[/], $other, 
                q[IL2/analysis/123456_IL2_1234/Data/Intensities/Bustard_RTA/GERALD_RTA/archive];
  mkdir join q[/], $archive, 'lane1';
  mkdir join q[/], $archive, 'lane2';
  mkdir join q[/], $archive, 'lane2', 'qc';
  mkdir join q[/], $archive, 'lane3';
  my $lqc = join q[/], $archive, 'lane3', 'qc';
  mkdir $lqc;
  my $file = join q[/], $archive, 'qc', '1234_3.insert_size.json';
  `cp $file $lqc`;
  mkdir join q[/], $archive, 'lane4';
  $lqc = join q[/], $archive, 'lane4', 'qc';
  mkdir $lqc;
  $file = join q[/], $archive, 'qc', '1234_4.insert_size.json';
  `cp $file $lqc`;

  local $ENV{TEST_DIR} = $temp;
  my $id_run = 1234;

  my $s = npg_qc::autoqc::qc_store->new(use_db    => q[no]);
  my $c = $s->load_run($id_run);
  is ($c->size, 16, 'loading main qc results only');
  
  $c = $s->load_run($id_run, 0, undef, $PLEXES);
  is ($c->size, 2, 'loading autoqc for plexes only');

  $c = $s->load_run($id_run, 0, [1,4,6], $PLEXES);
  is ($c->size, 1, 'loading autoqc for plexes only for 3 lanes');

  $c = $s->load_run($id_run, 0, [1,6], $PLEXES);
  is ($c->size, 0, 'loading autoqc for plexes only for 2 lanes, one empty, one no-existing');

  $c = $s->load_run($id_run, 0, [1,6], $ALL);
  is ($c->size, 4, 'loading all autoqc including plexes  for 2 lanes, for plexes one empty, one no-existing');

  $c = $s->load_run($id_run, 0,  [4], $ALL);
  is ($c->size, 3, 'loading all autoqc including plexes  for 1 lane');

  $c = $s->load_run($id_run, 0, [], $ALL);
  is ($c->size, 18, 'loading all autoqc');
}

1;